@@ -38,7 +38,7 @@
  *
  * The functions in this file refer to MRLs, which are a special form
  * of URIs used by xine to refer to things such as DVDs. An example of
- * an MRL would be <example>dvd:///dev/scd0</example>, which is not a
+ * an MRL would be <literal>dvd:///dev/scd0</literal>, which is not a
  * valid URI as far as, for example, GIO is concerned.
  *
  * The rest of the totem-pl-parser API exclusively uses URIs.
@@ -366,6 +366,7 @@ cd_cache_new (const char *dev,
       g_object_unref (mount);
     }
 
+    cache->content_types = g_content_type_guess_for_tree (cache->iso_file);
     cache->mountpoint = g_file_get_path (cache->iso_file);
     cache->mounted = TRUE;
 
